/*
/$OBJINST file en/decoder
*/

#include <stdio.h>
#include <limits.h>
#include <errno.h>
#include <dirent.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/unistd.h>

typedef unsigned char uint8_t;
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;
typedef unsigned long long uint64_t;

typedef uint8_t BYTE; /* 1-byte integer */
typedef uint16_t WORD; /* 2-byte integer */
typedef uint32_t DWORD; /* 4-byte integer */
typedef uint64_t QWORD; /* 8-byte integer */

static WORD get_WORD( BYTE* b );
static DWORD get_DWORD( BYTE* b );
static QWORD get_QWORD( BYTE* b );
static WORD fread_WORD( FILE* f );
static DWORD fread_DWORD( FILE* f );
static QWORD fread_QWORD( FILE* f );

BYTE* buf = NULL;
uint buflen = 0;
char name[1000];

int main( int argc, char* argv[] ){
	uint i;
	for( i = 1; i <= argc; i++ ){
		FILE* f;
		f = fopen( argv[i], "rb" );
		if( f ){
			DWORD num_entries, offset, length, entry;
			long cur_offset;

			fseek( f, 4, SEEK_SET );
			num_entries = fread_DWORD( f );

			for( entry = 0; entry < num_entries; entry++ ){
				offset = fread_DWORD( f );
				length = fread_DWORD( f );
				if( buflen < length ){
					BYTE* t = (BYTE*)realloc( buf, length+1 );
					if( !t ){
						printf( "ERROR: Not enough RAM\n" );
						return 1;
					}
					buf = t;
					buflen = length;
				}
				cur_offset = ftell( f );
				fseek( f, offset, SEEK_SET );
				fread( buf, length, 1, f );
				fseek( f, cur_offset, SEEK_SET );
				
				{
					sprintf( name, "%u", entry );
					mkdir( name, 0666 );
					chdir( name );
					{
						FILE* out = fopen( argv[i], "wb" );
						if(out){
							fwrite( buf, length, 1, out );
							fclose( out );
						}
					}
					chdir( ".." );
				}

			}
			fclose( f );
		}
	}
	return 0;
}

static WORD get_WORD( BYTE* b ){
	return b[0] | (b[1] << 8);
}

static DWORD get_DWORD( BYTE* b ){
	return b[0] | (b[1] << 8) | (b[2] << 16) | (b[3] << 24);
}

static QWORD get_QWORD( BYTE* b ){
	return b[0] | (b[1] << 8) | (b[2] << 16) | (b[3] << 24)
		| ((QWORD)b[4] << 32) | ((QWORD)b[5] << 40) | ((QWORD)b[6] << 48) | ((QWORD)b[7] << 56);
}

static WORD fread_WORD( FILE* f ){
	BYTE b[2]; fread( b, 2, 1, f );
	return b[0] | (b[1] << 8);
}

static DWORD fread_DWORD( FILE* f ){
	BYTE b[4]; fread( b, 4, 1, f );
	return b[0] | (b[1] << 8) | (b[2] << 16) | (b[3] << 24);
}

static QWORD fread_QWORD( FILE* f ){
	BYTE b[8]; fread( b, 8, 1, f );
	return b[0] | (b[1] << 8) | (b[2] << 16) | (b[3] << 24)
		| ((QWORD)b[4] << 32) | ((QWORD)b[5] << 40) | ((QWORD)b[6] << 48) | ((QWORD)b[7] << 56);
}